from time import time
import binascii
import hashlib
import json
from contextlib import contextmanager
from datetime import datetime
from json import JSONEncoder
from uuid import UUID
import shortuuid


class Event:
    """
    Event handlers storage.

    Example:

    >>> def handler(**kwargs):
    ...     print("Triggered:", kwargs)
    ...
    >>> Event.get('some_thing_happened').attach(handler)
    >>> Event.get('some_thing_happened').emit(status='OK')
    Triggered: {'status': 'OK'}

    """

    _repo = {}

    def __init__(self, name):
        self.handlers = []
        self.name = name

    def attach(self, handler, priority=0):
        handler._event_priority = priority
        self.handlers.append(handler)
        self.handlers.sort(key=lambda h: h._event_priority)

    def detach(self, handler):
        self.handlers.remove(handler)

    @classmethod
    def get(cls, name):
        instance = cls._repo.get(name, None)
        if not instance:
            instance = cls(name)
            cls._repo[name] = instance
        return instance

    @contextmanager
    def subscribe(self, *handlers):
        for handler in handlers:
            self.attach(handlers)
        try:
            yield
        finally:
            for handlers in handlers:
                self.detach(handlers)

    def emit(self, *args, **kwargs):
        for handler in self.handlers:
            handler(*args, **kwargs)


def on(event_name, priority=0):
    def decorator(fn):
        Event.get(event_name).attach(fn, priority)
    return decorator


class AttrDict(dict):
    def __init__(self, *args, **kwargs):
        dict.__init__(self, *args, **kwargs)
        self.__dict__ = self

class VersatileEncoder(JSONEncoder):
    """
    Encode all "difficult" object such as UUID
    """

    def _iterencode(self, obj, markers=None):

        if isinstance(obj, tuple) and hasattr(obj, '_asdict'):
            gen = self._iterencode_dict(obj._asdict(), markers)
        else:
            gen = JSONEncoder._iterencode(self, obj, markers)
        for chunk in gen:
            yield chunk

    def default(self, obj):
        if isinstance(obj, UUID):
            return str(obj)
        elif isinstance(obj, datetime):
            return str(obj)
        else:
            try:
                return JSONEncoder.default(self, obj)
            except TypeError:
                return str(obj)


def to_json(obj):
    return json.dumps(obj, ensure_ascii=False, cls=VersatileEncoder).replace("</", "<\\/")

def from_json(s):
    return AttrDict(json.loads(s))

def make_rand(length=16):
    return shortuuid.ShortUUID().random(length)

def make_hash(secret, as_binary=False, **kwargs):
    kwargs = kwargs or dict(
        salt=b'tokit',
        iterations=82016,
    )
    # might be GIL dependent
    dk = hashlib.pbkdf2_hmac('sha512', str.encode(secret), **kwargs)
    if as_binary:
        return dk
    else:
        return binascii.hexlify(dk).decode()

def md5(text):
    m = hashlib.md5()
    m.update(str.encode(text))
    return binascii.hexlify(m.digest()).decode()

try:
    import markdown


    def md_html(raw):
        """
        Convert Markdown format string to HTML
        WARNING: Be careful of leading spaces
        """
        html = markdown.markdown(raw.decode())
        return html
except:
    pass

class cached_property(object):
    """
    A property that is only computed once per instance and then replaces itself
    with an ordinary attribute. Deleting the attribute resets the property.
    Source: https://github.com/bottlepy/bottle/commit/fa7733e075da0d790d809aa3d2f53071897e6f76
    """  # noqa

    def __init__(self, func):
        self.__doc__ = getattr(func, '__doc__')
        self.func = func

    def __get__(self, obj, cls):
        if obj is None:
            return self
        value = obj.__dict__[self.func.__name__] = self.func(obj)
        return value

class cached_property_ttl(object):
    """
    A property that is only computed once per instance and then replaces itself
    with an ordinary attribute. Setting the ttl to a number expresses how long
    the property will last before being timed out.
    Source: https://github.com/pydanny/cached-property/blob/master/cached_property.py
    """

    def __init__(self, ttl=None):
        if callable(ttl):
            func = ttl
            ttl = None
        else:
            func = None
        self.ttl = ttl
        self._prepare_func(func)

    def __call__(self, func):
        self._prepare_func(func)
        return self

    def __get__(self, obj, cls):
        if obj is None:
            return self

        now = time()
        obj_dict = obj.__dict__
        name = self.__name__
        try:
            value, last_updated = obj_dict[name]
        except KeyError:
            pass
        else:
            ttl_expired = self.ttl and self.ttl < now - last_updated
            if not ttl_expired:
                return value

        value = self.func(obj)
        obj_dict[name] = (value, now)
        return value

    def __delete__(self, obj):
        obj.__dict__.pop(self.__name__, None)

    def __set__(self, obj, value):
        obj.__dict__[self.__name__] = (value, time())

    def _prepare_func(self, func):
        self.func = func
        if func:
            self.__doc__ = func.__doc__
            self.__name__ = func.__name__
            self.__module__ = func.__module__



from tokit import Request
from tokit.utils import on
from tokit.compiler import init_complier
from tokit.translation import init_locale, TranslationMixin

@on('init')
def init(app):
    init_complier(app)
    init_locale(app.config)

class Home(TranslationMixin, Request):

    URL = '/'

    def get(self):
        self.render('home.html')

    def css(self):
        yield 'base.css'
        yield 'home/home.sass'

    def js(self):
        yield from ['riot.js', 'home/home.coffee', 'home/home.tag']


class About(TranslationMixin, Request):

    URL = '/about', 'about'

    def get(self):
        self.render('about.jade')
        
    def css(self):
        yield 'base.css'



"""
Transform to browsers' languages
Can serve files directly using shortcut: `python3 -m tokit.compiler`

Require system's libsass and pyv8
"""
import os
import io
import tornado.ioloop
import tornado.web
from tornado.iostream import IOStream
from tornado.gen import coroutine
from tornado.web import HTTPError
from tokit.tasks import ThreadPoolMixin, run_on_executor
from tokit import ValidPathMixin

COMPILER_URLS = []

def read_file(filename):
    with io.open(filename, encoding='utf8') as fp:
        return fp.read()


class CompilerHandler(ThreadPoolMixin, ValidPathMixin, tornado.web.RequestHandler):

    def set_default_headers(self):
        self.set_header('Server', 'Python3')

    @coroutine
    def get(self, requested_file):
        try:
            requested_path = requested_file.replace(self.application.settings['static_url_prefix'], '')
            abs_path = os.path.abspath(os.path.join(self.application.root_path, requested_path))
            self.validate_absolute_path(self.application.root_path, abs_path)
            yield self.compile(abs_path)
        except Exception as e:
            self.set_status(400)
            self.write('/*')
            self.write(str(e))
            self.write('*/')

try:
    import sass

    class SassHandler(CompilerHandler):

        def prepare(self):
            self.set_header('Content-Type', 'text/css')

        @run_on_executor
        def compile(self, full_path):
            result = sass.compile(
                filename=full_path,
                output_style=('nested' if self.application.settings['debug'] else 'compressed')
            )
            self.write(result)

    COMPILER_URLS.append((r'^(/.+\.sass)$', SassHandler))
except ImportError:
    pass

try:
    import execjs
    js_context = execjs.get().compile(
        read_file(os.path.dirname(__file__) + '/js/coffee-script.js')
    )
    riot_context = execjs.get().compile(
        read_file(os.path.dirname(__file__) + '/js/coffee-script.js') +
        ";\n\n var exports = {}; module.exports = {};" +  # fake CommonJS environment
        read_file(os.path.dirname(__file__) + '/js/riotc.js') + "; var riot = module.exports;"
    )

    class CoffeeHandler(CompilerHandler):

        def prepare(self):
            self.set_header('Content-Type', 'application/javascript')

        @run_on_executor
        def compile(self, full_path):
            result = js_context.call(
                "CoffeeScript.compile",
                read_file(full_path),
                {'bare': True}
            )
            self.write(result)

    class RiotHandler(CoffeeHandler):

        @run_on_executor
        def compile(self, full_path):
            result = riot_context.call(
                "riot.compile",
                read_file(full_path),
                True
            )
            self.write(result)


    COMPILER_URLS.append((r'^(/.+\.coffee)$', CoffeeHandler))
    COMPILER_URLS.append((r'^(/.+\.tag)$', RiotHandler))
except ImportError:
    pass

def init_complier(app):
    from tokit import logger
    if len(COMPILER_URLS):
        app.add_handlers('.*$', COMPILER_URLS)
        app.root_path = app.config.root_path
    else:
        logger.warn('Found no compilers handlers')

if __name__ == "__main__":
    app = tornado.web.Application(urlspecs())
    app.root_path = os.path.realpath(os.path.curdir)
    print("Serving via compiler in:", app.root_path)
    app.listen(7998)
    tornado.ioloop.IOLoop.current().start()



from cgitb import *
from cgitb import __UNDEF__

def trace_html(einfo, context=5):
    """Return a nice HTML document describing a given traceback."""
    # src https://github.com/python/cpython/blob/3.6/Lib/cgitb.py#L101
    etype, evalue, etb = einfo
    if isinstance(etype, type):
        etype = etype.__name__
    pyver = 'Python ' + sys.version.split()[0] + ': ' + sys.executable
    date = time.ctime(time.time())
    head = '<html><head><style>body { margin: 0 } p { padding: 10px; } * { font-family: monospace; }  strong::before { content: " "; display: block; }</style></head><body bgcolor="#f0f0f8">' + pydoc.html.heading(
        '<big><big>%s</big></big>' %
        strong(pydoc.html.escape(str(etype))),
        '#ffffff', '#6622aa', pyver + '<br>' + date) + '''
<p>A problem occurred in a Python script.  Here is the sequence of
function calls leading up to the error, in the order they occurred.</p>'''

    indent = '<tt>' + small('&nbsp;' * 5) + '&nbsp;</tt>'
    frames = []
    records = inspect.getinnerframes(etb, context)
    for frame, file, lnum, func, lines, index in records:
        if file:
            file = os.path.abspath(file)
            link = '<a href="file://%s">%s</a>' % (file, pydoc.html.escape(file))
        else:
            file = link = '?'
        args, varargs, varkw, locals = inspect.getargvalues(frame)
        call = ''
        if func != '?':
            call = 'in ' + strong(func) + \
                inspect.formatargvalues(args, varargs, varkw, locals,
                    formatvalue=lambda value: '=' + pydoc.html.repr(value))

        highlight = {}
        def reader(lnum=[lnum]):
            highlight[lnum[0]] = 1
            try: return linecache.getline(file, lnum[0])
            finally: lnum[0] += 1
        vars = scanvars(reader, frame, locals)

        rows = ['<tr><td bgcolor="#d8bbff">%s%s %s</td></tr>' %
                ('<big>&nbsp;</big>', link, call)]
        if index is not None:
            i = lnum - index
            for line in lines:
                num = small('&nbsp;' * (5-len(str(i))) + str(i)) + '&nbsp;'
                if i in highlight:
                    line = '<tt>=&gt;%s%s</tt>' % (num, pydoc.html.preformat(line))
                    rows.append('<tr><td bgcolor="#ffccee">%s</td></tr>' % line)
                else:
                    line = '<tt>&nbsp;&nbsp;%s%s</tt>' % (num, pydoc.html.preformat(line))
                    rows.append('<tr><td>%s</td></tr>' % grey(line))
                i += 1

        done, dump = {}, []
        for name, where, value in vars:
            if name in done: continue
            done[name] = 1
            if value is not __UNDEF__:
                if where in ('global', 'builtin'):
                    name = ('<em>%s</em> ' % where) + strong(name)
                elif where == 'local':
                    name = strong(name)
                else:
                    name = where + strong(name.split('.')[-1])
                dump.append('%s&nbsp;= %s' % (name, pydoc.html.repr(value)))
            else:
                dump.append(name + ' <em>undefined</em>')

        rows.append('<tr><td>%s</td></tr>' % small(grey(', '.join(dump))))
        frames.append('''
<table width="100%%" cellspacing=0 cellpadding=0 border=0>
%s</table>''' % '\n'.join(rows))

    exception = ['<p>%s: %s' % (strong(pydoc.html.escape(str(etype))),
                                pydoc.html.escape(str(evalue)))]
    for name in dir(evalue):
        if name[:1] == '_': continue
        value = pydoc.html.repr(getattr(evalue, name))
        exception.append('\n<br>%s%s&nbsp;=\n%s' % (indent, name, value))

    ret = head + ''.join(exception) + ''.join(reversed(frames))
    return ret.replace('#6622aa', '#5f99cf') \
        .replace('#d8bbff', '#cee3f8') \
        .replace('#f0f0f8', '#fff') \
        .replace('#ffccee', '#e7e7e7') \
        .replace('#909090', '#000')



#!/usr/bin/env python3
import os
from tornado import options as opts
import tokit


def main():
    opts.define('host', default='::1')
    opts.define('port', default='9000')
    opts.define('env', default=None)
    opts.parse_command_line()

    config = tokit.Config(__file__)
    config.set_env(opts.options.env)

    tokit.start(opts.options.host, opts.options.port, config)

if __name__ == '__main__':
    main()



import logging
import shortuuid
import uuid

import momoko
from psycopg2.extras import DictCursor, DictRow, register_uuid
import psycopg2.extensions

from tornado.gen import coroutine

import tokit
from tokit import api

logger = tokit.logger

class DictLogCursor(DictCursor):

    def execute(self, sql, args=None):
        logger.debug('Excute SQL: %s', self.mogrify(sql, args).decode())
        return super().execute(sql, args)


@tokit.on('init')
def pg_init(app):
    """ Hook to init Postgres momoko driver.
    dsn config is required, with syntax same as Psycopg2 DSN.

    Sample env.ini::

        [postgres]
        dsn=dbname=[APP_NAME]
        size=2
    """
    env = app.config.env['postgres']
    if env.getboolean('log_momoko'):
        logging.getLogger('momoko').setLevel(logger.getEffectiveLevel())
    momoko_opts = dict(
        dsn=env['dsn'],
        size=int(env['size']),
        max_size=int(env['max_size']),
        auto_shrink=env.getboolean('auto_shrink'),
        cursor_factory=(DictLogCursor if env.getboolean('log') else DictCursor),
        # connection_factory=env.get('connection_factory', None),
    )
    register_uuid()
    app.pg_db = momoko.Pool(**momoko_opts)
    app.pg_db.connect()


class PgMixin:

    DbIntegrityError = psycopg2.IntegrityError
    DbError = psycopg2.Error

    @property
    def db(self):
        return self.application.pg_db

    @coroutine
    def pg_insert(self, table, fields=None, **data):
        """
        Postgres shorcut to insert data
        :return int new row's id

        Example::

            user_id = yield self.pg_insert('users', {"username": "foo", "password": "secret"})
        """
        if fields:
            data = self.get_request_dict(*fields)
        else:
            fields = list(data.keys())
        assert len(data) > 0  # check data
        values = list(data.values())

        sql = 'INSERT INTO {} ({}) VALUES ({}) RETURNING id ' \
            .format(table,
                    ','.join(fields),
                    ','.join(['%s'] * len(fields))
                    )
        cursor = yield self.db.execute(sql, values)
        return cursor.fetchone()[0]

    @coroutine
    def pg_update(self, table, data):
        changes = [field + ' = %s' for field in data.keys()]
        sql = 'UPDATE {} SET {} WHERE id = %s'.format(table, ','.join(changes))
        values = list(data.values()) + [data['id']]
        yield self.db.execute(sql, values)

    @coroutine
    def pg_query(self, query, *params):
        """ Low level execuation """
        result = yield self.db.execute(query, params)
        return result

    def pg_serialize(self, row):
        if not row:
            return
        ret = dict(row) if isinstance(row, DictRow) else row
        return ret

    @coroutine
    def pg_select(self, query, *params):
        """
        Query and convert each returned row

        :return generator
        """
        result = yield self.pg_query(query, *params)
        return (self.pg_serialize(row) for row in result.fetchall())

    @coroutine
    def pg_one(self, query, *params):
        result = yield self.pg_query(query, *params)
        row = result.fetchone()
        if row:
            return self.pg_serialize(row)

    db_insert = pg_insert
    db_update = pg_update
    db_query = pg_query
    db_select = pg_select
    db_one = pg_one


class UidMixin:

    def pg_serialize(self, row):
        ret = PgMixin.pg_serialize(self, row)
        if 'id' in ret:
            ret['short_id'] = shortuuid.encode(ret['id'])
        return ret



import re
import os
from collections import ChainMap, defaultdict

from tornado.template import (
    Loader, ParseError,
    Template as TornadoTemplate
)
try:
    from pyjade.ext.tornado import Template as JadeTemplate
except:
    pass

from tornado import locale
from tokit.utils import on



SHORTCUT_RE = [
    (re.compile(rb'{\*\s*([\w\_]+)\s+(.*?)\*}', re.DOTALL), rb'{{ _("\1").format(\2) }}'),
    (re.compile(rb'{\*\s*([\w\_]+)\s*\*}', re.DOTALL),              rb'{{ _("\1") }}'),
]

def init_locale(config):
    """
    Load per-module `lang` folder CSV translations
    """
    chain = defaultdict(lambda: defaultdict(ChainMap))

    for m in config.modules_loaded:
        lang_path = os.path.join(config.root_path, m, 'lang')
        if not os.path.exists(lang_path):
            continue
        locale.load_translations(lang_path)

        # old translations is overrided by a `load` call,
        # so this is a hack to chain all mobule's translations into bigger one
        for lang, plurals in locale._translations.items():
            for plural, translation in plurals.items():
                chain[lang][plural].maps.append(translation)

    locale._translations = chain
    locale._supported_locales = frozenset(list(chain.keys()) + [locale._default_locale])


class CustomLoader(Loader):
    """
    Tornado's template preprocessor with translation shortcut.

        * ``{* key | name=value *}``-> ``{{ _("key").format(name=value) }}``
        * ``{* key *}``-> ``{{ _("key") }}``
    """

    def _custom_prepocessor(self, content):
        ret = content
        for regex, replacement in SHORTCUT_RE:
            ret = regex.sub(replacement, ret)
        return ret

    def _create_template(self, name):
        path = os.path.join(self.root, name)
        with open(path, "rb") as f:
            try:
                EngineClass = self._get_template_engine(name)
                template = EngineClass(
                    self._custom_prepocessor(f.read()),
                    name=name, loader=self,
                    compress_whitespace=False
                )
                return template
            except ParseError as exception:
                exception.args = exception.args + (path, )
                raise
                
    def _get_template_engine(self, name):
        if name.endswith('.jade'):
            return JadeTemplate

        return TornadoTemplate


class TranslationMixin:

    def create_template_loader(self, template_path):
        settings = self.application.settings
        if "template_loader" in settings:
            return settings["template_loader"]
        kwargs = {}
        if "autoescape" in settings:
            # autoescape=None means "no escaping", so we have to be sure
            # to only pass this kwarg if the user asked for it.
            kwargs["autoescape"] = settings["autoescape"]
        return CustomLoader(template_path, **kwargs)



import os
from setuptools import setup
from itertools import chain

# Thanks to http://peterdowns.com/posts/first-time-with-pypi.html
tokit_version = '0.7'

def pack_files(directory):
    for (path, directories, filenames) in os.walk(directory):
        for filename in filenames:
            yield os.path.join(path, filename)

extra_files = list(chain(pack_files('tokit/js'), pack_files('tokit/skeleton')))

setup(
    name='tokit',
    version=tokit_version,
    description='A kit for development with Tornado web framework',
    author='Giang Manh',
    author_email='manhgd@yahoo.com',
    url='https://github.com/manhg/tokit',
    download_url='https://github.com/manhg/tokit/tarball/' + tokit_version,
    keywords=['tornado', 'web', 'tokit'],
    install_requires = [
        "tornado==4.4",
        "shortuuid==0.4.3",
        "cerberus==1.0.1"
    ],
    packages=['tokit'],
    package_data={
        '': extra_files
    },
    classifiers=[
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ]
)
